import os
from pathlib import Path
# from dotenv import load_dotenv, find_dotenv, DotEnv
from dotenv import load_dotenv, find_dotenv, dotenv_values
dotenv_path = find_dotenv()
load_dotenv(dotenv_path)
# dotenv_vals = load_dotenv(dotenv_path)

REQUIRED_VARS = [
]

OPTIONAL_VARS = [
    'MYSQL_HOST',
    'MYSQL_USERNAME',
    'MYSQL_PASSWORD',
    'MYSQL_DEFAULT_DB',
    'MYSQL_ENCODING', # will default to "utf8mb4"
    'MYSQL_PROTOCOL', # will default to "mysql+pymysql"
]

class Config(object):

    def __init__(self, spark_mem=None, path_to_paper_data=None, path_to_citation_data=None):
        self.PROJECT_DIR = os.environ.get('PROJECT_DIR') or Path(__file__).resolve().parents[1]

        for var in REQUIRED_VARS:
            setattr(self, var, os.environ[var])

        for var in OPTIONAL_VARS:
            setattr(self, var, os.environ.get(var, default=None))
            
        self.PATH_TO_PAPER_DATA = path_to_paper_data or os.environ.get('DEFAULT_PAPER_DATA', "")
        self.PATH_TO_CITATION_DATA = path_to_citation_data or os.environ.get('DEFAULT_PAPER_CITATION_DATA', "")

        self._mysql_db = None
        self._spark = None
        self.spark_mem = spark_mem or os.environ.get('SPARK_MEM') or '80g'

        self._elasticsearch = None

    def get(self, x, default=None):
        return getattr(self, x, default)

    @property
    def mysql_db(self):
        """database object used for connecting to MySQL"""
        # if not hasattr(self, '_mysql') or self._mysql_db is None:
        if self.get('_mysql_db') is None:
            db_name = self.MYSQL_DEFAULT_DB or None
            self._mysql_db = self._get_mysql_connection(db_name=db_name)
        return self._mysql_db

    @mysql_db.deleter
    def mysql_db(self):
        del self._mysql_db

    @property
    def spark(self):
        if self.get('_spark') is None:
            self._spark = self.load_spark_session(mem=self.spark_mem)
        return self._spark

    @spark.deleter
    def spark(self):
        if self.get('_spark') is not None:
            self._spark.stop()
        del self._spark

    def restart_spark(self):
        del self.spark
        return self.spark

    @property
    def elasticsearch(self):
        if self.get('_elasticsearch') is None:
            self._elasticsearch = self.setup_elasticsearch()
        return self._elasticsearch
    
    def setup_elasticsearch(self, host='localhost', port=9200):
        from elasticsearch import Elasticsearch
        return Elasticsearch([{'host': host, 'port': port}])




    def _get_mysql_connection(self, db_name=None):
        """Use SQLAlchemy to get a connection to MySQL

        :db_name: optional name of database to connect to
        :returns: db object

        """
        mysql_settings = {
            'db_name': db_name,
            'host_name': self.MYSQL_HOST or '127.0.0.1',
            'user_name': self.MYSQL_USERNAME,
            'password': self.MYSQL_PASSWORD,
            'encoding': self.MYSQL_ENCODING or 'utf8mb4'
        }
        c = MySQLConnect(**mysql_settings)
        return c

    def reset_mysql_db(self, db_name=None):
        """Reset the MySQL DB Connection

        :db_name: optional name of database to connect to
        :returns: db object

        """
        self._mysql_db = self._get_mysql_connection(db_name=db_name)

    def load_spark_session(self, appName="sparkApp", mem='80g', showConsoleProgress=False, additional_conf=[], logLevel=None):
        # import findspark
        # findspark.init()

        import pyspark
        conf = pyspark.SparkConf().setAll([
            ('spark.executor.memory', mem), 
            ('spark.driver.memory', mem),
            ('spark.ui.showConsoleProgress', showConsoleProgress),
            ('spark.driver.maxResultSize', '0'),
            ('spark.reducer.maxSizeInFlight', '5g'),
            ("spark.sql.execution.arrow.enabled", "false"),
            ("spark.driver.extraJavaOptions", "-Duser.timezone=UTC"),  # https://stackoverflow.com/a/48767250
            ("spark.executor.extraJavaOptions", "-Duser.timezone=UTC"),
        ])
        for k,v in additional_conf:
            conf.set(k, v)

        sc = pyspark.SparkContext(appName=appName, conf=conf)
        if logLevel:
            sc.setLogLevel(logLevel)
        spark = pyspark.sql.SparkSession(sc)
        return spark

    def teardown(self):
        del self.spark
        del self.mysql_db
        

class MySQLConnect(object):

    """Use SQLAlchemy to connect to MySQL"""

    def __init__(self,
                 protocol = 'mysql+pymysql',
                 db_name = None,
                 host_name = None,
                 user_name = None,
                 password = None,
                 encoding = None,
                 module_path = None,
                 test = False):

        from sqlalchemy import create_engine, MetaData, Table
        self.create_engine = create_engine
        self.MetaData = MetaData
        self.Table = Table

        self.engine = None
        self.metadata = None
        self.tables = None

        self.db_name = db_name
        self.protocol = protocol.lower()
        self.host_name = host_name
        self.user_name = user_name
        self.password = password
        self.encoding = encoding
        self.module_path = module_path
        self.test = test

        if self.host_name:
            self.get_engine()

    def get_connection_string(self):
        con = self.protocol + "://"
        con = con + self.user_name
        if self.password:
            con = con + ":" + self.password
        con = con + "@" + self.host_name
        if self.db_name:
            con = con + "/" + self.db_name
        else:
            con = con + "/"
        con = con + "?local_infile=1"
        if self.encoding:
            con = con + "&charset=" + self.encoding
        return con

    def get_engine(self, pool_recycle=25200):
        """Get sqlalchemy engine
        :returns: sqlalchemy engine object

        """
        if not self.engine:
            self.engine = self.create_engine(self.get_connection_string(), pool_recycle=pool_recycle)
        if self.db_name:
            self.metadata = self.MetaData(self.engine)
            self.metadata.reflect()
            self.tables = self.get_all_tables()
        return self.engine

    def load_table(self, table_name=None):
        """Load a database table by name

        :table_name: database table name
        :returns: sqlalchemy Table object

        """
        return self.Table(table_name, self.metadata, autoload=True)

    def get_all_tables(self):
        """
        Returns a dictionary of {tablename: Table object}
        for the tables in the database
        """
        return {tablename: self.load_table(tablename) for tablename in self.engine.table_names()}

    # def read_sql(self, sq):
    #     """
    #     Use pandas.read_sql() to execute a SQL query to get a dataframe
    #     See pandas documentation
    #     """
    #     try:
    #         import pandas as pd
    #     except ImportError:
    #         raise ImportError("You need to install Pandas for this")
    #     return pd.read_sql(sq, self.engine)


    def __repr__(self):
        result = "DBConfiguration():\n\t"
        result = result + "PROTOCOL='%s'\n\t"%(self.protocol)
        result = result + "DATABASE_NAME='%s'\n\t"%(self.db_name)
        result = result + "DATABASE_HOST='%s'\n\t"%(self.host_name)
        result = result + "DATABASE_USER='%s'\n\t"%(self.user_name)
        if( self.password ):
            result = result + "PASSWORD <is_set>\n\t"
        else:
            result = result + "PASSWORD <is_NOT_set>\n\t"
        result = result + "MODULE_PATH='%s'"%(self.module_path)
        return result

    




# -*- coding: utf-8 -*-

DESCRIPTION = """Pajek Tools"""

import sys, os, time
from pathlib import Path
from typing import Optional
from datetime import datetime
from timeit import default_timer as timer

try:
    from humanfriendly import format_timespan
except ImportError:

    def format_timespan(seconds):
        return "{:.2f} seconds".format(seconds)


import logging

# logging.basicConfig(format='%(asctime)s %(name)s.%(lineno)d %(levelname)s : %(message)s',
#         datefmt="%H:%M:%S",
#         level=logging.INFO)
# logger = logging.getLogger(__name__)
# logger = logging.getLogger('__main__').getChild(__name__)
logger = logging.getLogger(__name__)
# logger.addHandler(logging.NullHandler())

import pandas as pd
import numpy as np
from csv import QUOTE_NONE


class PajekWriter:

    """Take a network as an edgelist and output a Pajek (.net) file"""

    def __init__(
        self,
        edgelist: pd.DataFrame,
        weighted: bool = False,
        vertices_label: str = "Vertices",
        edges_label: Optional[str] = None,
        directed: bool = True,
        citing_colname: str = "ID",
        cited_colname: str = "cited_ID",
        weight_colname: str = "weight",
    ):
        """

        Parameters
        ----------
        edgelist : pandas dataframe
            Should have a column for citing node name and cited node name.
            Optional column for weight.
        weighted : `bool`, optional
            Is this a network with weighted edges?
        vertices_label : `str`, default: "Vertices"
            label to use for the vertices
        edges_label : `str`, optional
            label to use for the edges. If None, "Arcs" will be used for
            directed networks, and "Edges" for undirected
        directed : `bool`, default: True
            Is this a directed network?
        citing_colname : `str`, default: "ID"
            Column label for the citing node name
        cited_colname : `str`, default: "cited_ID"
            Column label for the cited node name
        weight_colname : `str`, default: "weight"
            Column label for the edge weight, if this is a weighted network

        """
        self.df_edgelist = edgelist
        self.weighted = weighted
        self.vertices_label = vertices_label
        self.directed = directed
        if edges_label:
            self.edges_label = edges_label
        else:
            if self.directed:
                self.edges_label = "Arcs"
            else:
                self.edges_label = "Edges"
        self.citing_colname = citing_colname
        self.cited_colname = cited_colname
        self.weight_colname = weight_colname

        self.df_edgelist[citing_colname] = self.df_edgelist[citing_colname].astype(str)
        self.df_edgelist[cited_colname] = self.df_edgelist[cited_colname].astype(str)

        self.df_vertices = None
        self.id_map = None

        logger.debug("PajekWriter initialized")

    @property
    def df_edgelist(self):
        """
        edgelist : pandas dataframe
            Should have a column for citing node name and cited node name.
            Optional column for weight.
        self.df_edgelist = self.

        """
        return self._df_edgelist

    @df_edgelist.setter
    def df_edgelist(self, value):
        self._df_edgelist = value
        if self._df_edgelist is not None:
            self.num_edges = len(self.df_edgelist)

    @property
    def df_vertices(self):
        """dataframe of unique node names and assigned integer node IDs"""
        return self._df_vertices

    @df_vertices.setter
    def df_vertices(self, value):
        self._df_vertices = value
        if self._df_vertices is not None:
            self.num_vertices = len(self.df_vertices)

    def get_df_vertices(self):
        """Get a dataframe of unique node names and assinged integer node IDs

        Returns
        -------
        Pandas dataframe

        """
        logger.debug("getting vertices dataframe...")
        x = np.concatenate(
            (
                self.df_edgelist[self.citing_colname],
                self.df_edgelist[self.cited_colname],
            ),
            axis=0,
        )
        x = np.unique(x)
        df_vertices = pd.DataFrame(x, columns=["node_name"])
        df_vertices["node_id"] = range(1, len(df_vertices) + 1)
        df_vertices["node_name"] = df_vertices["node_name"].astype(str)
        self.df_vertices = df_vertices
        return self.df_vertices

    def get_id_map(self, df_vertices):
        """Get a pandas Series mapping node name to assigned integer node ID

        Parameters
        ----------
        df_vertices : Pandas DataFrame
            dataframe with 'node_name' and 'node_id' columns

        Returns
        -------
        Pandas Series

        """
        logger.debug("getting ID map...")
        self.id_map = df_vertices.set_index("node_name")["node_id"]
        return self.id_map

    def write(self, outf, chunksize: int = 10000000):
        """Write the network to a Pajek (.net) file

        Parameters
        ----------
        outf : str or Path
            path to output file (will be overwritten if exists)

        """
        if self.id_map is None:
            if self.df_vertices is None:
                self.df_vertices = self.get_df_vertices()
            self.id_map = self.get_id_map(self.df_vertices)
        self.df_edgelist["citing_id"] = self.df_edgelist[self.citing_colname].map(
            self.id_map
        )
        self.df_edgelist["cited_id"] = self.df_edgelist[self.cited_colname].map(
            self.id_map
        )
        outfpath = Path(outf)
        quotechar = '"'
        logger.debug("opening output file: {}".format(outfpath))
        with open(outfpath, "w", newline="\n") as outfile:
            logger.debug("writing {} vertices...".format(self.num_vertices))
            outfile.write("*{} {}\n".format(self.vertices_label, self.num_vertices))
            self.df_vertices["node_name"] = (
                quotechar + self.df_vertices["node_name"] + quotechar
            )
            self.df_vertices[["node_id", "node_name"]].to_csv(
                outfile,
                sep=" ",
                index=False,
                header=False,
                quoting=QUOTE_NONE,
                line_terminator="\n",
                chunksize=chunksize,
            )

            logger.debug("writing {} edges...".format(self.num_edges))
            outfile.write("*{} {}\n".format(self.edges_label, self.num_edges))

            outcols = ["citing_id", "cited_id"]
            if self.weighted:
                outcols.append(self.weight_colname)
            self.df_edgelist[outcols].to_csv(
                outfile,
                sep=" ",
                index=False,
                header=False,
                quoting=QUOTE_NONE,
                line_terminator="\n",
                chunksize=chunksize,
            )


def main(args):
    pass


if __name__ == "__main__":
    total_start = timer()
    logger.info(" ".join(sys.argv))
    logger.info("{:%Y-%m-%d %H:%M:%S}".format(datetime.now()))
    import argparse

    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser.add_argument("--debug", action="store_true", help="output debugging info")
    global args
    args = parser.parse_args()
    if args.debug:
        logger.setLevel(logging.DEBUG)
        logger.debug("debug mode is on")
    else:
        logger.setLevel(logging.INFO)
    main(args)
    total_end = timer()
    logger.info(
        "all finished. total time: {}".format(format_timespan(total_end - total_start))
    )
